"""backfillz namespace."""

from importlib_metadata import PackageNotFoundError, version

from backfillz.backfillz import Backfillz

# Used to automatically set version number from github actions
# as well as not break when being tested locally
try:
    __version__ = version(__package__)  # type: ignore
except PackageNotFoundError:  # pragma: no cover
    __version__ = "0.0.0"



from dataclasses import dataclass
from math import floor
from typing import Any, Dict, List

import numpy as np
from stan.fit import Fit  # type: ignore


@dataclass
class MCMCRun:
    """A Stan fit, plus some derived data."""

    samples: Fit

    def iter_chains(self, param: str, index: int = 0) -> np.ndarray:
        """Return (n_chains × n_samples) matrix of draws for a given parameter."""
        n_chains, n_samples = self.samples.num_chains, self.samples.num_samples
        xss = np.zeros((n_chains, n_samples))
        for n in range(0, n_chains):
            xss[n] = self.samples[param][index][n * n_samples: (n + 1) * n_samples]
        return xss

    @property
    def params(self) -> List[str]:
        return list(self.samples.param_names)


@dataclass
class Slice:
    """A slice of an MCMC trace."""

    lower: float
    upper: float


Param = str
Slices = Dict[Param, List[Slice]]
Props = Dict[str, Any]


@dataclass
class ParameterSlices:
    """The MCMC data being presented."""

    slcs: List[Slice]
    param: str
    chains: np.ndarray  # shape is [n, n_iter] where n is number of chains
    max_sample: float
    min_sample: float

    @property
    def n_iter(self) -> int:
        """Return number of MCMC iterations per chain."""
        return int(self.chains.shape[1])

    def chain_slices(self, slc: Slice) -> List[np.ndarray]:
        """The specified slice of each chain."""
        return [
            self.chains[
                n,
                floor(slc.lower * self.n_iter):floor(slc.upper * self.n_iter)
            ]
            for n, _ in enumerate(self.chains)
        ]



from dataclasses import dataclass
import math
from typing import cast, List, Sequence, Tuple

import numpy as np
from plotly.basedatatypes import BaseTraceType  # type: ignore
import plotly.graph_objects as go  # type: ignore

from backfillz.data import MCMCRun, ParameterSlices, Props, Slice
from backfillz.plot import alpha, annotate, LeafPlot, Plot, segment, VerticalSubplots
from backfillz.slice_histograms import SliceHistogram
from backfillz.theme import BackfillzTheme


@dataclass
class DialPlot(LeafPlot):
    """Trace dial plot on the left."""

    hole_size: float = 1 / 3
    donut_domain: Tuple[float, float] = 0.5 * math.pi, 2 * math.pi

    @property
    def plot_elements(self) -> List[BaseTraceType]:
        return [self.donut] + [trace for trace in self.polar_traces]  # type conversion

    @property
    def xaxis_props(self) -> Props:
        # scaleanchor='y', scaleratio=1 would force an exact square/circle, but then can't
        # set histogram widths correctly
        return dict(visible=False)

    @property
    def yaxis_props(self) -> Props:
        return dict(range=[-1, 1], visible=False)

    @staticmethod
    def to_angular(x: float, domain: Tuple[float, float]) -> float:
        """Normalised x coordinate as angular coordinate in specified portion of unit circle."""
        start, end = domain
        return start + x * (end - start)

    @staticmethod
    def to_radial(y: float) -> float:
        """Map a normalised y coordinate into upper 2/3 of radius."""
        return DialPlot.hole_size + y * (1 - DialPlot.hole_size)

    # Bit inefficient for chains (we compute min/max rather than used the cached property on self.data).
    @staticmethod
    def normalise(xs: Sequence[float]) -> List[float]:
        min_x: float = min(xs)
        max_x: float = max(xs)
        return [(x - min_x) / (max_x - min_x) for x in xs]

    @staticmethod
    def polar_plot(xs: List[float], ys: List[float]) -> Tuple[List[float], List[float]]:
        xs_ang = [DialPlot.to_angular(x, DialPlot.donut_domain) for x in xs]
        return ([math.cos(x) * ys[n] for n, x in enumerate(xs_ang)],
                [math.sin(x) * ys[n] for n, x in enumerate(xs_ang)])

    def polar_trace(self, n: int) -> go.Scatter:
        chain = [(x, y) for x, y in enumerate(self.data.chains[n])]
        xs = DialPlot.normalise([x for x, _ in chain])
        ys = [DialPlot.to_radial(y) for y in DialPlot.normalise([y for _, y in chain])]
        xs_circ, ys_circ = DialPlot.polar_plot(xs, ys)
        return go.Scatter(
            x=xs_circ, y=ys_circ,
            line=dict(color=self.theme.palette[n]),
        )

    @property
    def donut(self) -> go.Scatter:
        n_segments: int = 100
        xs1 = [0] + [*range(0, n_segments)]
        ys1 = [DialPlot.hole_size] + [1.0] * n_segments
        assert len(xs1) == len(ys1)
        xs2 = [n_segments - 1] + [*range(n_segments - 1, -1, -1)]
        ys2 = [1.0] + [DialPlot.hole_size] * n_segments
        assert len(xs2) == len(ys2)
        xs_circ, ys_circ = DialPlot.polar_plot(DialPlot.normalise(xs1 + xs2), ys1 + ys2)
        return go.Scatter(
            x=xs_circ, y=ys_circ,
            line=dict(width=0),
            fill='toself',
            fillcolor=self.theme.mg_colour,
        )

    @property
    def polar_traces(self) -> List[go.Scatter]:
        return [self.polar_trace(n) for n, _ in enumerate(self.data.chains)]


@dataclass
class TraceDialHistogram(SliceHistogram):
    """Slice histogram for trace dial plot."""

    @property
    def plot_elements(self) -> List[BaseTraceType]:
        return [self.histo([n], self.theme.palette[n], 1) for n, _ in enumerate(self.data.chains)]

    @property
    def xaxis_props(self) -> Props:
        if self.n_slc == len(self.data.slcs) - 1:
            return dict(side='top')
        else:
            return dict(visible=False)


@dataclass
class SliceHistograms(VerticalSubplots):
    """One slice histogram per slice."""

    def make_plots(self) -> List[Plot]:
        return [
            TraceDialHistogram(
                axis_id=self.axis_ids[n],
                x_domain=self.x_domain,
                y_domain=segment(self.y_domain, len(self.data.slcs), n),
                data=self.data,
                theme=self.theme,
                slc=slc,
                n_slc=n,
                row=self.row + len(self.data.slcs) - 1 - n,
                col=self.col,
            )
            for n, slc in enumerate(self.data.slcs)
        ]


# TODO: consolidate with plot.RootPlot
@dataclass
class TraceDial:
    """Top-level plot, for a given parameter and chain."""

    data: ParameterSlices
    theme: BackfillzTheme
    verbose: bool

    @property
    def plots(self) -> List[Plot]:
        return [self.dial_plot, self.histograms]

    @property
    def dial_plot(self) -> DialPlot:
        return DialPlot(
            axis_id='',
            x_domain=(0.0, 1),
            y_domain=(0.0, 1),
            row=1,
            col=1,
            data=self.data,
            theme=self.theme,
        )

    @property
    def histograms(self) -> SliceHistograms:
        return SliceHistograms(
            axis_ids=['3', '2'],
            x_domain=(0.5 + DialPlot.hole_size / 2, 1.0),
            y_domain=(0.5, 1.0),
            row=1,
            col=1,
            data=self.data,
            theme=self.theme,
        )

    @property
    def title(self) -> str:
        return f"Pretzel plot for {self.data.param}"

    def render(self) -> go.Figure:
        """Create fig and render subplots."""
        layout = go.Layout(
            title=self.title,
            titlefont=dict(size=30),
            plot_bgcolor=self.theme.bg_colour,
            showlegend=False,
            barmode='overlay',
            xaxis2=dict(anchor='y2'),
            yaxis2=dict(anchor='x2'),
            xaxis3=dict(anchor='y3'),
            yaxis3=dict(anchor='x3'),
            # plotting region won't be exactly square but best we can do to align histogram width with donut
            width=800, height=800,
        )
        fig = go.Figure(layout=layout)

        for plot in self.plots:
            plot.layout_axes(fig)

        self.add_additional_titles(fig)

        for trace in self.dial_plot.plot_elements:
            fig.add_trace(trace)

        for histo in self.histograms.plots:
            for trace in cast(LeafPlot, histo).plot_elements:
                fig.add_trace(trace)

        return fig

    def add_additional_titles(self, fig: go.Figure) -> None:
        histos: List[Plot] = self.histograms.plots
        annotate(fig, 14, histos[0].top_left, 'right', 'top', None, "Burn-in histogram", textangle=-90)
        annotate(fig, 14, histos[1].top_left, 'right', 'top', None, "Sample histogram", textangle=-90)

    @staticmethod
    def fig(mcmc_run: MCMCRun, theme: BackfillzTheme, verbose: bool, param: str) -> go.Figure:
        """Create a trace slice histogram."""
        slcs: List[Slice] = [Slice(0.0, 0.04), Slice(0.4, 1)]  # how to decide
        return TraceDial(ParameterSlices(
            slcs=slcs,
            param=param,
            chains=mcmc_run.iter_chains(param),
            max_sample=np.amax(mcmc_run.samples[param]),
            min_sample=np.amin(mcmc_run.samples[param]),
        ), theme, verbose).render()


# Not using these properties yet.
@dataclass
class DerivativeColours:
    """Colours uniquely determined by a theme."""

    theme: BackfillzTheme

    @property
    def trace_line(self) -> str:
        return self.theme.text_font_colour

    @property
    def guide_lines(self) -> str:
        return self.theme.fg_colour

    @property
    def inner_burn_segment(self) -> str:
        return alpha(self.theme.mg_colour, self.theme.alpha + 0.2)

    @property
    def outer_burn_segment(self) -> str:
        return alpha(self.theme.mg_colour, self.theme.alpha + 0.1)

    @property
    def remaining_segment(self) -> str:
        return alpha(self.theme.mg_colour, self.theme.alpha - 0.3)



from dataclasses import dataclass, field
from typing import Any, Dict, List, Literal, Optional, Set, Tuple

from plotly.basedatatypes import BaseTraceType  # type: ignore
from plotly.colors import unlabel_rgb  # type: ignore
import plotly.graph_objects as go  # type: ignore
from plotly.subplots import make_subplots  # type: ignore

from backfillz.data import ParameterSlices, Props
from backfillz.theme import BackfillzTheme


class AbstractMethodError(NotImplementedError):
    """MyPy doesn't support abstract data classes yet (https://github.com/python/mypy/issues/5374)."""

    pass


# strings assigned as axis id suffixes by Plotly; empty for first subplot
AxisId = str
# Plotly subplot specs; 2D array of dictionaries
Specs = List[List[object]]


def default_config() -> Props:
    """Preferred settings for Plotly figure."""
    return dict(displayModeBar=False, showAxisDragHandles=False)


def cols(xss: Specs) -> int:
    """Length of the inner lists."""
    ns: Set[int] = set(map(len, xss))
    assert len(ns) == 1
    return min(ns)


def scale(factor: float, xs: List[float]) -> List[float]:
    """Element-wise product."""
    return [x * factor for x in xs]


def segment(domain: Tuple[float, float], n: int, m: int) -> Tuple[float, float]:
    """Break supplied "domain" into n equal-sized segments, and return the mth."""
    start, end = domain
    width = (end - start) / n
    return start + m * width, start + (m + 1) * width


def alpha(colour: str, a: float) -> str:
    """Add an alpha component to a colour represented as an RGB string."""
    rgb: tuple[int, int, int] = unlabel_rgb(colour)
    return f"rgb({rgb[0]},{rgb[1]},{rgb[2]},{a})"


def annotate(
    fig: go.Figure,
    font_size: int,
    at: Tuple[float, float],
    xanchor: Literal['left', 'right'],
    yanchor: Literal['top', 'bottom'],
    y_adjust: Optional[float],  # additional normalised offet of text relative to plot
    text: str,
    textangle: int = 0,
) -> None:
    """Add an annotation to supplied figure, with supplied arguments in addition to some default settings."""
    fig.add_annotation(
        xref='paper',
        yref='paper',
        showarrow=False,
        font=dict(size=font_size),
        x=at[0],
        y=at[1] + (0 if y_adjust is None else y_adjust),
        xanchor=xanchor,
        yanchor=yanchor,
        text=text,
        textangle=textangle
    )


@dataclass
class Plot:
    """Base class providing common subplot functionality."""

    x_domain: Tuple[float, float]  # left/right edges normalised to [0, 1]
    y_domain: Tuple[float, float]  # top/bottom edges normalised to [0, 1]
    row: int
    col: int
    data: ParameterSlices
    theme: BackfillzTheme

    def layout_axes(self, fig: go.Figure) -> None:
        raise AbstractMethodError()

    def render(self, fig: go.Figure) -> None:
        """Render me into fig."""
        raise AbstractMethodError()

    @property
    def top_left(self) -> Tuple[float, float]:
        return self.x_domain[0], self.y_domain[1]


@dataclass
class LeafPlot(Plot):
    """A leaf subplot."""

    # Axis ids (and annotation ids) need hand-configuration to match assignment by Plotly.
    axis_id: AxisId

    @property
    def plot_elements(self) -> List[BaseTraceType]:
        raise AbstractMethodError()

    def render(self, fig: go.Figure) -> None:
        for el in self.plot_elements:
            fig.add_trace(el, self.row, self.col)

    @property
    def axis_defaults(self) -> Dict[str, Any]:
        return dict(
            showgrid=False,
            zeroline=False,
            linecolor=self.theme.fg_colour,
            ticks='outside',
            tickwidth=1,
            ticklen=5,
            tickcolor=self.theme.fg_colour,
            fixedrange=True,  # disable selection zoom
        )

    @property
    def xaxis_id(self) -> str:
        """My Plotly-assigned x-axis id."""
        return 'xaxis' + self.axis_id

    @property
    def yaxis_id(self) -> str:
        """My Plotly-assigned y-axis id."""
        return 'yaxis' + self.axis_id

    def layout_axes(self, fig: go.Figure) -> None:
        """Configure my x and y axis settings in fig."""
        fig.layout[self.xaxis_id].update(domain=self.x_domain, **self.axis_defaults, **self.xaxis_props)
        fig.layout[self.yaxis_id].update(domain=self.y_domain, **self.axis_defaults, **self.yaxis_props)

    @property
    def xaxis_props(self) -> Props:
        """My custom x-axis settings; subclasses can override."""
        return dict()

    @property
    def yaxis_props(self) -> Props:
        """My custom y-axis settings; subclasses can override."""
        return dict()


@dataclass
class LeafPlotNoAxes(Plot):
    """A leaf subplot of "domain" type, i.e. with no axes."""

    @property
    def plot_elements(self) -> List[BaseTraceType]:
        raise AbstractMethodError()

    def render(self, fig: go.Figure) -> None:
        for el in self.plot_elements:
            fig.add_trace(el)

    def layout_axes(self, fig: go.Figure) -> None:
        pass


@dataclass
class VerticalSubplots(Plot):
    """A collection of vertically arranged subplots."""

    axis_ids: List[AxisId]  # one per subplots
    plots: List[Plot] = field(init=False)

    def __post_init__(self) -> None:
        self.plots = self.make_plots()

    def make_plots(self) -> List[Plot]:
        """My subplots."""
        raise AbstractMethodError()

    def layout_axes(self, fig: go.Figure) -> None:
        """Ask each subplot to configure its axes."""
        for plot in self.plots:
            plot.layout_axes(fig)

    def render(self, fig: go.Figure) -> None:
        """Render subplots into fig."""
        for n, plot in enumerate(self.plots):
            plot.render(fig)


# Should consolidate some of the commonality with Plot.
@dataclass
class RootPlot:
    """Top-level plot container."""

    theme: BackfillzTheme
    verbose: bool

    @property
    def plots(self) -> List[Plot]:
        raise AbstractMethodError()

    def grid_specs(self, fig: go.Figure) -> Specs:
        raise AbstractMethodError()

    @property
    def title(self) -> str:
        """Title for overall figure."""
        raise AbstractMethodError()

    def add_additional_titles(self, fig: go.Figure) -> None:
        annotate(
            fig, 14, (1, -0.03), 'right', 'top', None,  # leave room for an x-axis, if needed
            "Backfillz-py by CIM, University of Warwick and The Alan Turing Institute"
        )

    def render(self) -> go.Figure:
        """Create fig and render subplots."""
        fig: go.Figure = go.Figure(
            layout=go.Layout(
                title=self.title,
                titlefont=dict(size=30),
                plot_bgcolor=self.theme.bg_colour,
                showlegend=False,
            )
        )

        specs: Specs = self.grid_specs(fig)

        make_subplots(
            rows=len(specs),
            cols=cols(specs),
            figure=fig,
            specs=specs,
            horizontal_spacing=0,
            vertical_spacing=0,
            print_grid=self.verbose,
        )

        for plot in self.plots:
            plot.layout_axes(fig)

        self.add_additional_titles(fig)

        for plot in self.plots:
            plot.render(fig)

        return fig



from dataclasses import dataclass
from typing import List

import numpy as np
from plotly.basedatatypes import BaseTraceType  # type: ignore
import plotly.graph_objects as go  # type: ignore

from backfillz.data import MCMCRun, ParameterSlices, Props, Slice
from backfillz.plot import annotate, LeafPlot, Plot, RootPlot, scale, segment, Specs, VerticalSubplots
from backfillz.slice_histograms import SliceHistogram
from backfillz.theme import BackfillzTheme


@dataclass
class TracePlot(LeafPlot):
    """Left-hand component."""

    @property
    def plot_elements(self) -> List[BaseTraceType]:
        return self.traces + self.boxes

    # one per chain
    @property
    def traces(self) -> List[go.Scatter]:
        return [
            go.Scatter(
                x=chain,
                y=list(range(0, self.data.n_iter)),
                line=dict(color=self.theme.palette[n])
            )
            for n, chain in enumerate(self.data.chains)
        ]

    # one per slice
    @property
    def boxes(self) -> List[go.Scatter]:
        return [
            go.Scatter(
                x=[self.data.min_sample] * 2 + [self.data.max_sample] * 2 + [self.data.min_sample],
                y=scale(self.data.n_iter, [slc.lower, slc.upper, slc.upper, slc.lower, slc.lower]),
                mode='lines',
                line=dict(width=2, color=self.theme.fg_colour),
            )
            for slc in self.data.slcs
        ]

    @property
    def xaxis_props(self) -> Props:
        return dict(range=[self.data.min_sample, self.data.max_sample])

    @property
    def yaxis_props(self) -> Props:
        return dict(range=[0, self.data.n_iter])


@dataclass
class JoiningSegments(LeafPlot):
    """Middle component."""

    @property
    def plot_elements(self) -> List[BaseTraceType]:
        return self.segments + [self.y_labels]

    # one per slice
    @property
    def segments(self) -> List[go.Scatter]:
        return [
            go.Scatter(
                x=[0, 1, 1, 0],
                y=scale(self.data.n_iter, [slc.lower, lower, upper, slc.upper]),
                mode='lines',
                line=dict(color=self.theme.fg_colour, width=1),
                fill='toself',
                fillcolor='rgba(240,240,240,255)'
            )
            for n, slc in enumerate(self.data.slcs, start=1)
            for lower, upper in [((n - 1) / len(self.data.slcs), n / len(self.data.slcs))]
        ]

    # one numerical marker per slice delimiter
    @property
    def y_labels(self) -> go.Scatter:
        y = self.slice_delimiters
        return go.Scatter(
            x=[0] * len(y),
            y=y,
            mode='text',
            text=[int(y) for y in y],
            textposition='middle right'
        )

    @property
    def slice_delimiters(self) -> List[float]:
        """Unique slice start/end points, expressed in iterations."""
        delims: List[float] = [*{*[y for slc in self.data.slcs for y in [slc.lower, slc.upper]]}]
        return scale(self.data.n_iter, delims)

    @property
    def xaxis_props(self) -> Props:
        return dict(rangemode='nonnegative', visible=False)

    @property
    def yaxis_props(self) -> Props:
        return dict(
            range=[0, self.data.n_iter],
            tickmode='array',
            tickvals=self.slice_delimiters,
            showticklabels=False
        )


@dataclass
class SliceHistograms(VerticalSubplots):
    """One slice histogram per slice."""

    def make_plots(self) -> List[Plot]:
        return [
            SliceHistogram(
                axis_id=self.axis_ids[n],
                x_domain=self.x_domain,
                y_domain=segment(self.y_domain, len(self.data.slcs), n),
                data=self.data,
                theme=self.theme,
                slc=slc,
                n_slc=n,
                row=self.row + len(self.data.slcs) - 1 - n,
                col=self.col,
            )
            for n, slc in enumerate(self.data.slcs)
        ]


@dataclass
class TraceSliceHistogram(RootPlot):
    """Trace slice histogram plot for a given parameter."""

    data: ParameterSlices
    left_w = 0.4  # width of trace plot
    middle_w = 0.2  # width of joining segments

    @property
    def plots(self) -> List[Plot]:
        return [self.trace_plot, self.joining_segments, self.density_plots]

    @property
    def trace_plot(self) -> TracePlot:
        return TracePlot(
            axis_id='',
            x_domain=(0, self.left_w),
            y_domain=(0, 1.0),
            row=1,
            col=1,
            data=self.data,
            theme=self.theme,
        )

    @property
    def joining_segments(self) -> JoiningSegments:
        return JoiningSegments(
            axis_id='2',
            x_domain=(self.left_w, self.left_w + self.middle_w),
            y_domain=(0, 1.0),
            row=1,
            col=2,
            data=self.data,
            theme=self.theme,
        )

    @property
    def density_plots(self) -> SliceHistograms:
        return SliceHistograms(
            axis_ids=[str(n + 3) for n in reversed(range(0, len(self.data.slcs)))],
            x_domain=(self.left_w + self.middle_w, 1),
            y_domain=(0, 1.0),
            row=1,
            col=3,
            data=self.data,
            theme=self.theme,
        )

    def grid_specs(self, fig: go.Figure) -> Specs:
        return (
            [[dict(rowspan=len(self.data.slcs)), dict(rowspan=len(self.data.slcs)), dict()]] +
            [[None, None, dict()] for _ in self.data.slcs[1:]]
        )

    @property
    def title(self) -> str:
        return f"Trace slice histogram of {self.data.param}"

    def add_additional_titles(self, fig: go.Figure) -> None:
        super().add_additional_titles(fig)
        annotate(fig, 16, self.trace_plot.top_left, 'left', 'bottom', None, "Trace Plot With Slices")
        # oof -- adjust for x-axis
        annotate(fig, 16, self.density_plots.top_left, 'left', 'bottom', 0.03, "Density Plots for Slices")

    @staticmethod
    def fig(mcmc_run: MCMCRun, theme: BackfillzTheme, verbose: bool, param: str) -> go.Figure:
        """Create a slice histogram."""
        slcs: List[Slice] = [Slice(0.028, 0.04), Slice(0.1, 0.2), Slice(0.4, 0.9)]
        return TraceSliceHistogram(theme, verbose, ParameterSlices(
            slcs=slcs,
            param=param,
            chains=mcmc_run.iter_chains(param),
            max_sample=np.amax(mcmc_run.samples[param]),
            min_sample=np.amin(mcmc_run.samples[param]),
        )).render()



from __future__ import annotations

import stan  # type: ignore

from backfillz.stan import Stan


schools_code = """
data {
  int<lower=0> J;         // number of schools
  real y[J];              // estimated treatment effects
  real<lower=0> sigma[J]; // standard error of effect estimates
}
parameters {
  real mu;                // population treatment effect
  real<lower=0> tau;      // standard deviation in treatment effects
  vector[J] eta;          // unscaled deviation from mu by school
}
transformed parameters {
  vector[J] theta = mu + tau * eta;        // school treatment effects
}
model {
  target += normal_lpdf(eta | 0, 1);       // prior log-density
  target += normal_lpdf(y | theta, sigma); // log-likelihood
}
"""

schools_data = {"J": 8,
                "y": [28, 8, -3, 7, -1, 1, 18, 12],
                "sigma": [15, 10, 16, 11, 9, 11, 10, 18]}


def generate_fit() -> Stan:
    """Make PyStan fit from the Eight Schools example."""
    model = stan.build(schools_code, data=schools_data, random_seed=1)
    return Stan(model, model.sample(num_chains=4, num_samples=12000))


if __name__ == '__main__':
    generate_fit()



from dataclasses import dataclass
from typing import List


@dataclass
class BackfillzTheme:
    """Backfillz visualisation settings."""

    name: str
    text_family: str
    text_font: float
    text_font_colour: str
    text_cex_title: float
    text_cex_main: float
    text_cex_axis: float
    text_col_title: str
    text_col_main: str
    text_col_axis: str
    bg_colour: str
    mg_colour: str
    fg_colour: str
    alpha: float
    palette: List[str]


default: BackfillzTheme = BackfillzTheme(
    name="default",
    text_family="sans",
    text_font=1,
    text_font_colour="black",
    text_cex_title=1.5,
    text_cex_main=1,
    text_cex_axis=0.8,
    text_col_title="#1a1a1a",
    text_col_main="#999999",
    text_col_axis="#666666",
    bg_colour="white",
    mg_colour="#7f7f7f",
    fg_colour="black",
    alpha=0.7,
    palette=[
        "#FF0000",
        "#0000FF",
        "#FF00FF",
        "#800000",
        "#000080",
        "#FF6347"
    ]
)

solarized_dark: BackfillzTheme = BackfillzTheme(
    name="solarized_dark",
    text_family="mono",
    text_font=1,
    text_font_colour="#2AA198",
    text_cex_title=2,
    text_cex_main=1,
    text_cex_axis=0.8,
    text_col_title="#1a1a1a",
    text_col_main="#999999",
    text_col_axis="#666666",
    bg_colour="#002B36",
    mg_colour="#7f7f7f",
    fg_colour="#93A1A1",
    alpha=0.7,
    palette=[
        "#657B83",
        "#D30102",
        "#D33682",
        "#859900",
        "#93A1A1",
        "#268BD2"
    ]
)

demo_1: BackfillzTheme = BackfillzTheme(
    name="demo 1",
    text_family="mono",
    text_font=1,
    text_font_colour="#333333",
    text_cex_title=1.5,
    text_cex_main=1,
    text_cex_axis=0.6,
    text_col_title="#1a1a1a",
    text_col_main="#999999",
    text_col_axis="#666666",
    bg_colour="#fafafa",
    mg_colour="#e5e5e5",
    fg_colour="#666666",
    alpha=0.8,
    palette=[
        "#A3C96D",
        "#DDCF1E",
        "#8E4D91",
        "#003B24",
        "#912B2F",
        "#7C6EAC"
    ]
)

demo_2: BackfillzTheme = BackfillzTheme(
    name="demo 2",
    text_family="sans",
    text_font=1,
    text_font_colour="#e5e5e5",
    text_cex_title=1.5,
    text_cex_main=1,
    text_cex_axis=0.6,
    text_col_title="#F2EEE7",
    text_col_main="#F2EEE7",
    text_col_axis="#F2EEE7",
    bg_colour="#313C3F",
    mg_colour="#313C3F",  # adjustcolor("#313C3F", red.f = 1.8, green.f = 1.8, blue.f = 1.8),
    fg_colour="#F2EEE7",
    alpha=0.8,
    palette=[
        "#EEE436",
        "#00AEC7",
        "#C73475",
        "#7FC5D3",
        "#7EB627",
        "#F29530"
    ]
)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['backfillz', 'backfillz.example']

package_data = \
{'': ['*']}

install_requires = \
['importlib_metadata==3.4.0',
 'numpy==1.20.3',
 'plotly>=4.14.3,<5.0.0',
 'pystan==3.0.0',
 'scipy>=1.6.3,<2.0.0']

setup_kwargs = {
    'name': 'backfillz',
    'version': '0.1.2',
    'description': 'MCMC visualisations package developed at the University of Warwick and supported by The Alan Turing Institute.',
    'long_description': '<!-- badges: start -->\n\n[![Release build](https://github.com/WarwickCIM/backfillz-py/actions/workflows/build-publish.yml/badge.svg?branch=release)](https://github.com/WarwickCIM/backfillz-py/actions/workflows/build-publish.yml)\n[![Develop build](https://github.com/WarwickCIM/backfillz-py/actions/workflows/build-publish.yml/badge.svg?branch=develop)](https://github.com/WarwickCIM/backfillz-py/actions/workflows/build-publish.yml)\n[![Project Status: WIP – Initial development is in progress, but there has not yet been a stable, usable release suitable for the public.](https://www.repostatus.org/badges/latest/wip.svg)](https://www.repostatus.org/#wip)\n[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)\n<!-- badges: end -->\n\n<img src="https://github.com/WarwickCIM/backfillz/raw/master/fig1.png" width=100% alt=""/>\n\n# New View of MCMC\n\nBackfillz-py provides new visual diagnostics for understanding MCMC (Markov Chain Monte Carlo) analyses and outputs. MCMC chains can defy a simple line graph. Unless the chain is very short (which isn’t often the case), plotting tens or hundreds of thousands of data points reveals very little other than a ‘trace plot’ where we only see the outermost points. Common plotting methods may only reveal when an MCMC really hasn’t worked, but not when it has.\nBackFillz-py slices and dices MCMC chains so increasingly parameter rich, complex analyses can be visualised meaningfully. What does ‘good mixing’ look like? Is a ‘hair caterpillar’ test verifiable? What does a density plot show and what does it hide?\n\n# Quick Start\n\nInstall from [PyPI](https://pypi.org/project/backfillz/).\n\nTODO - code example\n\n# Current prototype plots\n\nTODO\n\n# Acknowledgements\n\nWe are grateful for funding from the Alan Turing Institute within the ‘Tools, Practices and Systems’ theme. Initial user research was carried out by GJM on the ‘2020 Science’ programme (www.2020science.net/) funded by the EPSRC Cross-Discipline Interface Programme (grant number EP/I017909/1).\n',
    'author': 'James Tripp',
    'author_email': 'james.tripp@warwick.ac.uk',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/WarwickCIM/backfillz-py',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9.0,<3.10',
}


setup(**setup_kwargs)



from dataclasses import dataclass
from math import ceil, floor
from typing import List

import numpy as np
from plotly.basedatatypes import BaseTraceType  # type: ignore
import plotly.graph_objects as go  # type: ignore
import scipy.stats as stats  # type: ignore

from backfillz.data import Props, Slice
from backfillz.plot import LeafPlot


@dataclass
class SliceHistogram(LeafPlot):
    """Plot histograms for arbitrary subsets of chains, plus optional KDE plots for individual chains."""

    slc: Slice
    n_slc: int

    @property
    def plot_elements(self) -> List[BaseTraceType]:
        """Histogram for a slice (aggregating all chains) plus density plot for each chain."""
        ns: List[int] = [n for n, _ in enumerate(self.data.chains)]
        return [self.histo(ns, self.theme.fg_colour, 1)] + [self.chain_plot(n) for n in ns]

    # Histogram for any subset of the chains.
    def histo(self, ns: List[int], color: str, bin_size: float) -> go.Histogram:
        chain_slices: List[np.ndarray] = self.data.chain_slices(self.slc)
        return go.Histogram(
            x=[x for n in ns for x in chain_slices[n]],
            xbins=dict(start=floor(self.data.min_sample), end=ceil(self.data.max_sample), size=bin_size),
            marker=dict(
                color=self.theme.bg_colour,
                line=dict(color=color, width=1)
            ),
            histnorm='probability',
            xaxis='x' + self.axis_id,
            yaxis='y' + self.axis_id,
        )

    # Non-parametric KDE, smoothed with a Gaussian kernel, for a given chain.
    def chain_plot(self, n: int) -> go.Scatter:
        x = np.linspace(self.data.min_sample, self.data.max_sample, 200)
        chain_slices = self.data.chain_slices(self.slc)
        return go.Scatter(
            x=x,
            y=stats.kde.gaussian_kde(chain_slices[n])(x),
            mode='lines',
            line=dict(width=2, color=self.theme.palette[n]),
            xaxis='x' + self.axis_id,
            yaxis='y' + self.axis_id,
        )

    @property
    def xaxis_props(self) -> Props:
        bottom: bool = self.n_slc == 0
        top: bool = self.n_slc == len(self.data.slcs) - 1
        # single slice requires special treatment; haven't figured out how to mirror tick labels
        if len(self.data.slcs) == 1:
            return dict(mirror='ticks')
        elif bottom:
            return dict()
        elif top:
            return dict(side='top')
        else:
            return dict(visible=False)

    @property
    def yaxis_props(self) -> Props:
        return dict(side='right', rangemode='nonnegative')
